/*
    Copyright (C) 2005 Wolfgang Hottgenroth

    This file is part of smmapdfw.

    smmapdfw is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.  

    smmapdfw is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
    License for more details. 

    You should have received a copy of the GNU General Public License
    along with smmapdfw. If not, write to the Free Software Foundation, 59
    Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


#if HAVE_CONFIG_H
# include "config.h"
#endif

#include <string.h>
#include <sys/types.h>
#include <assert.h>
#include <syslog.h>

#include "htmalloc.h"
#include "htbuffer.h"



#define MAX(A, B)  ((A) > (B) ? (A) : (B))
#define MIN(A, B)  ((A) < (B) ? (A) : (B))

htbuffer_t *htbuffer_init(unsigned int start_size) {
  htbuffer_t *b;

  b = (htbuffer_t*) htmalloc(sizeof(htbuffer_t));

  b->current_buf_size = (start_size % DEFAULT_CHUNK_SIZE) 
    ? (start_size + (DEFAULT_CHUNK_SIZE - start_size % DEFAULT_CHUNK_SIZE)) 
    : start_size;
    
  b->max_size = DEFAULT_MAX_SIZE;

  b->buf = (char*) htmalloc(sizeof(char) * b->current_buf_size);
  memset(b->buf, 0, b->current_buf_size);

  return b;
}

void htbuffer_set_max_size(htbuffer_t *b, unsigned int s) {
  b->max_size = s;
}

void htbuffer_free(htbuffer_t *b) {
  free(b->buf);
  free(b);
}
    

unsigned int htbuffer_strlen(htbuffer_t *b) {
  return strlen(b->buf);
}

htbuffer_t *htbuffer_memcpy_w_offset(htbuffer_t *dest, unsigned int offset, const char *src, unsigned int src_size) {
  int new_buf_size;

/*   printf("src_size=%d, offset=%d, cbs=%d\n", src_size, offset, dest->current_buf_size); */
  if (src_size + offset > dest->current_buf_size) {
    /* grow buffer */
    int pos = src_size + offset;
    new_buf_size = (pos % DEFAULT_CHUNK_SIZE) 
      ? (pos + (DEFAULT_CHUNK_SIZE - (pos % DEFAULT_CHUNK_SIZE))) 
      : pos;

    if (new_buf_size > dest->max_size) {
      syslog(LOG_ERR, "htbuffer (%p) htbuffer_memcpy_w_offset: new size too large %d (max=%d)",
	     dest, new_buf_size, dest->max_size);
      exit(154);
    }
    
    syslog(LOG_INFO, "htbuffer (%p) htbuffer_memcpy_w_offset: increase buffer from %d to %d\n",
	   dest, dest->current_buf_size, new_buf_size);
    dest->buf = (char*) htrealloc(dest->buf, sizeof(char) * new_buf_size);
    memset(dest->buf + dest->current_buf_size, 0, new_buf_size - dest->current_buf_size);
    dest->current_buf_size = new_buf_size;
  }

  memcpy(dest->buf + offset, src, src_size);

  return dest;
}


htbuffer_t *htbuffer_clear(htbuffer_t *dest) {
  memset(dest->buf, 0, dest->current_buf_size);
  return dest;
}

htbuffer_t *htbuffer_memcpy(htbuffer_t *dest, const char *src, unsigned int src_size) {
  return htbuffer_memcpy_w_offset(dest, 0, src, src_size);
}

htbuffer_t *htbuffer_strcpy_w_offset(htbuffer_t *dest, unsigned int offset, const char *src) {
  return htbuffer_memcpy_w_offset(dest, offset, src, strlen(src)+1);
}

htbuffer_t *htbuffer_strcpy(htbuffer_t *dest, const char *src) {
  return htbuffer_strcpy_w_offset(dest, 0, src);
}

htbuffer_t *htbuffer_strncpy_w_offset(htbuffer_t *dest, unsigned int offset, const char *src, unsigned int n) {
  return htbuffer_memcpy_w_offset(dest, offset, src, MIN(strlen(src)+1, n));
}

htbuffer_t *htbuffer_strncpy(htbuffer_t *dest, const char *src, unsigned int n) {
  return htbuffer_strncpy_w_offset(dest, 0, src, n);
}

htbuffer_t *htbuffer_strcat(htbuffer_t *dest, const char *src) {
  return htbuffer_strcpy_w_offset(dest, strlen(dest->buf), src);
}

htbuffer_t *htbuffer_strncat(htbuffer_t *dest, const char *src, unsigned int n) {
  return htbuffer_strncpy_w_offset(dest, strlen(dest->buf), src, n);
}







  



